#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ctype.h>

#include "netboggle.h"
#include "trie.h"


const char *wordracer2 =
"..XX.."
".XXXX."
"XXXXXX"
"XXXXXX"
".XXXX."
"..XX..";

const char *wordracer3 =
"XXXX.."
"XXXX.."
"XXXXXX"
"XXXXXX"
"..XXXX"
"..XXXX";

const char *wordracer4 =
"XXXXXX"
"XXXXXX"
"XX..XX"
"XX..XX"
"XXXXXX"
"XXXXXX";



/* set the "mode" of the board to be either normal boggle, or wordracer */
void set_mode(char *mode, int *board_size,
              int graph[MAX_BOX_SIZE][MAX_BOX_SIZE])
{
  int x, y, i;
  const char *board_template = NULL;

  if (!strcmp(mode, "wr2")) board_template = wordracer2;
  else if (!strcmp(mode, "wr3")) board_template = wordracer3;
  else if (!strcmp(mode, "wr4")) board_template = wordracer4;

  if (board_template == NULL) {
    *board_size = 0;
    for (y = 0; y < MAX_BOX_SIZE; y++)
      for (x = 0; x < MAX_BOX_SIZE; x++)
        graph[x][y] = NOT_SEEN;
  } else {
    *board_size = 6;
    i = 0;
    for (y = 0; y < 6; y++)
      for (x = 0; x < 6; x++)
        graph[x][y] = (board_template[i++] == 'X' ? NOT_SEEN : NOT_USED);
  }
}



/* Read the board from input.  Return the dimension of the board. */
void read_board(char *raw_input, int letters[MAX_BOX_SIZE][MAX_BOX_SIZE],
                int graph[MAX_BOX_SIZE][MAX_BOX_SIZE],
                int *board_size)
{
  int i, j, num, len, let_number;

  len = strlen(raw_input);

  /* if this is a wordracer board (i.e. one with a board template),
   * make sure we have the correct number of letters */
  if (*board_size) {
    num = 0;
    for (j = 0; j < *board_size; j++)
      for (i = 0; i < *board_size; i++)
        if (graph[i][j] == NOT_SEEN)
          num++;
    if (num != len) {
      printf("Illegal board!  The board shape you've selected requires %d\n"
             "letters.  You only provided %d.\n", num, len);
      exit(1);
    }

  /* if this is not a wordracer board, dynamically determine the size */
  } else {
    for (i = 2; i <= MAX_BOX_SIZE; i++)
      if (len == i*i) {
        *board_size = i;
        break;
      }

    if (*board_size == 0) {
      printf("Illegal board size!  The board must be square, but you entered\n"
             "%d letters.  Largest allowed board size is %d x %d.\n", len,
             MAX_BOX_SIZE, MAX_BOX_SIZE);
      exit(1);
    }
  }

  /* put the letters into their proper matrix positions */
  num = 0;
  for (j = 0; j < *board_size; j++) {
    for (i = 0; i < *board_size; i++) {

      /* if this is an unused space (i.e. in a wordracer board), skip it */
      if (graph[i][j] == NOT_USED)
        continue;

      /* convert the letter into its corresponding number */
      if (raw_input[num] == 'Q')
        let_number = QU;
      else
        let_number = tolower(raw_input[num]) - 'a';

      if (let_number != QU && (let_number < 0 || let_number >= NUM_LETTERS)) {
        printf("illegal character '%c' in board\n", raw_input[num]);
        exit(1);
      } else {
        letters[i][j] = let_number;
      }
      num++;
    }
  }
}



/*************************************************************************/


int get_char(int fd)
{
  char c;

  if (read(fd, &c, sizeof(char)) != sizeof(char))
    return BOG_EOF;

  if (c == '%')
    return BOG_EOF;

  if (c == '\n' || c == '\r')
    return BOG_EOLN;

  return c;
}


int get_line(char *buf, int fd, int maxsize)
{
  int i, c;

  i = 0;
  while (i < maxsize-1 && (c = get_char(fd)) != BOG_EOF) {
    /* end of line character: ignore if it's at the beginning, otherwise
       end the line */
    if (c == BOG_EOLN) {
      if (i)
        break;
      else
        continue;
    }
    buf[i++] = c;
  }

  buf[i] = '\0';
  return i;
}
    


void read_config(int fd, int letters[MAX_BOX_SIZE][MAX_BOX_SIZE],
                 int graph[MAX_BOX_SIZE][MAX_BOX_SIZE],
                 int *board_size, int *min_length, int *format)
{
  char buf[TMPBUFSIZE], *attribute, *value, *eq, *p;
  int have_board = 0, have_mode = 0;

  /* initialize the default values we want */
  *min_length = 0;
  *format = FORMAT_PLAIN;

  /* now read each config line off the network... */
  while (get_line(buf, fd, TMPBUFSIZE)) {

    /* parse the line for attribute and value, separated by '=' */
    attribute = value = eq = NULL;
    for (p = buf; *p != '\0'; p++) {
      if (attribute == NULL && isalnum(*p))
        attribute = p;
      if (*p == '=') {
        *p = '\0';
        eq = p;
      }
      if (eq != NULL && isalnum(*p)) {
        value = p;
        break;
      }
    }

    if (attribute == NULL || value == NULL) {
      fprintf(stderr, "invalid line\n");
      continue;
    }

    /* now we have attribute and value... parse the contents */

    /* the "minength" directive */
    if (!strcmp(attribute, "minlength"))
      *min_length = atoi(value);

    /* the "mode" directive */
    else if (!strcmp(attribute, "mode")) {
      if (have_board) {
        printf("mode directive must come before board!\n");
        exit(1);
      } else {
        set_mode(value, board_size, graph);
        have_mode = 1;
      }

    /* the "board" directive */
    } else if (!strcmp(attribute, "board")) {
      /* the mode must be defined first... */
      if (!have_mode)
        set_mode("normal", board_size, graph);
      read_board(value, letters, graph, board_size);
      have_board = 1;

    /* the "format" directive */
    } else if (!strcmp(attribute, "format")) {
    }

    /* default - no known directive */
    } else {
      printf("unknown keyword %s!\n", attribute);
      exit(1);
    }
  }

  if (!have_board) {
    printf("No board provided!\n");
    exit(1);
  }

  /* if no min length is explicitly provided, the default is one less
   * than the board dimension */
  if (*min_length == 0) {
    *min_length = *board_size - 1;
  }
}

